"""Numba-optimized functions for calculating various statistics."""

import numpy as np
from numba import njit


@njit(cache=True)
def col_cumsum(x: np.ndarray):
    """Calculate the cumulative sum of each column in a 2D array.

    Parameters
    ----------
    x : np.ndarray
        2D array.
    """
    cumsum = np.zeros_like(x)
    for j in range(x.shape[1]):
        cumsum[:, j] = np.cumsum(x[:, j])
    return cumsum


@njit(cache=True)
def log_det_covariance(X: np.ndarray) -> float:
    """Compute log determinant of the covariance matrix of a data matrix.

    Parameters
    ----------
    X : np.ndarray
        2D array of shape (n, p) where n is the number of samples and p is the number of
        variables.

    Returns
    -------
    log_abs_det : float
        The log of the absolute value of the determinant of the covariance matrix.
        Returns np.nan if the covariance matrix is not positive definite.

    """
    p = X.shape[1]
    cov = np.cov(X, rowvar=False, ddof=0).reshape(p, p)
    det_sign, log_abs_det = np.linalg.slogdet(cov)

    if det_sign <= 0:
        return np.nan
    else:
        return log_abs_det



"""Interactive exploration of the Capa and Mvcapa anomaly detectors."""

import pandas as pd
import plotly.express as px

from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.mvcapa import Mvcapa
from skchange.datasets.generate import generate_teeth_data
from skchange.utils.benchmarking.profiler import Profiler

# Unviariate
df = generate_teeth_data(n_segments=5, segment_length=10, mean=10, random_state=2)[0]
detector = Capa(max_segment_length=20)

anomalies = detector.fit_predict(df)
print(anomalies)

anomaly_labels = detector.fit_transform(df)
px.scatter(x=df.index, y=df, color=anomaly_labels.astype(str))

scores = detector.score_transform(df)
px.scatter(scores)

# Multivariate
df = generate_teeth_data(5, 10, p=10, mean=10, affected_proportion=0.2, random_state=2)
detector = Mvcapa(collective_penalty="sparse")

anomalies = detector.fit_predict(df)
print(anomalies)

anomaly_labels = detector.fit_transform(df)
anomaly_labels = (anomaly_labels > 0).astype(int)
anomaly_labels[anomaly_labels == 0] = 0.1
plot_df = pd.concat(
    [
        df.melt(ignore_index=False).reset_index(),
        anomaly_labels.melt(value_name="anomaly_label")["anomaly_label"],
    ],
    axis=1,
)
plot_df["variable"] = plot_df["variable"].astype(str)
px.scatter(plot_df, x="index", y="value", color="variable", size="anomaly_label")

fig = px.line(df)
fig.add_scatter(anomaly_labels)
px.line(anomaly_labels)

scores = detector.score_transform(df)
px.scatter(scores)


# Profiling
n = int(1e5)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = Capa(
    max_segment_length=100, collective_penalty_scale=5, point_penalty_scale=5
)
detector = Mvcapa(
    max_segment_length=1000,
    collective_penalty="sparse",
    collective_penalty_scale=5,
    point_penalty_scale=5,
)
profiler = Profiler().start()
detector.fit_predict(df)
profiler.stop()



"""Base classes for changepoint detectors."""

import numpy as np
import pandas as pd

from skchange.base import BaseDetector


class ChangeDetector(BaseDetector):
    """Base class for changepoint detectors.

    Changepoint detectors detect points in time where a change in the data occurs.
    Data between two changepoints is a segment where the data is considered to be
    homogeneous, i.e., of the same distribution. A changepoint is defined as the
    location of the last element of a segment.

    Output format of the predict method: See the dense_to_sparse method.
    Output format of the transform method: See the sparse_to_dense method.

    Subclasses should set the following tags for sktime compatibility:
    - task: "change_point_detection"
    - learning_type: "unsupervised" or "supervised"
    - And possibly other tags, such as
        * "capability:missing_values": False,
        * "capability:multivariate": True,
        * "fit_is_empty": False,

    Needs to be implemented:
    - _fit(self, X, y=None) -> self
    - _predict(self, X) -> pd.Series

    Optional to implement:
    - _score_transform(self, X) -> pd.Series
    - _update(self, X, y=None) -> self
    """

    @staticmethod
    def sparse_to_dense(
        y_sparse: pd.Series, index: pd.Index, columns: pd.Index = None
    ) -> pd.Series:
        """Convert the sparse output from the predict method to a dense format.

        Parameters
        ----------
        y_sparse : pd.DataFrame
            The sparse output from a changepoint detector's predict method.
        index : array-like
            Indices that are to be annotated according to ``y_sparse``.
        columns: array-like
            Not used. Only for API compatibility.

        Returns
        -------
        pd.Series with integer labels 0, ..., K for each segment between two
            changepoints.
        """
        changepoints = y_sparse.to_list()
        n = len(index)
        changepoints = [-1] + changepoints + [n - 1]
        segment_labels = np.zeros(n)
        for i in range(len(changepoints) - 1):
            segment_labels[changepoints[i] + 1 : changepoints[i + 1] + 1] = i

        return pd.Series(
            segment_labels, index=index, name="segment_label", dtype="int64"
        )

    @staticmethod
    def dense_to_sparse(y_dense: pd.Series) -> pd.Series:
        """Convert the dense output from the transform method to a sparse format.

        Parameters
        ----------
        y_dense : pd.Series
            The dense output from a changepoint detector's transform method.

        Returns
        -------
        pd.Series of changepoint locations. Changepoints are defined as the last element
            of a segment.
        """
        y_dense = y_dense.reset_index(drop=True)
        # changepoint = end of segment, so the label diffs > 0 must be shiftet by -1.
        is_changepoint = np.roll(y_dense.diff().abs() > 0, -1)
        changepoints = y_dense.index[is_changepoint]
        return ChangeDetector._format_sparse_output(changepoints)

    @staticmethod
    def _format_sparse_output(changepoints) -> pd.Series:
        """Format the sparse output of changepoint detectors.

        Can be reused by subclasses to format the output of the _predict method.
        """
        return pd.Series(changepoints, name="changepoint", dtype="int64")



"""Tests for all annotators/detectors in skchange."""

import numpy as np
import pandas as pd
import pytest
from sktime.utils._testing.annotation import make_annotation_problem
from sktime.utils.estimator_checks import check_estimator, parametrize_with_checks

from skchange.anomaly_detectors import ANOMALY_DETECTORS
from skchange.base import BaseDetector
from skchange.change_detectors import CHANGE_DETECTORS
from skchange.datasets.generate import generate_anomalous_data

ALL_DETECTORS = ANOMALY_DETECTORS + CHANGE_DETECTORS


@parametrize_with_checks(ALL_DETECTORS)
def test_sktime_compatible_estimators(obj, test_name):
    check_estimator(obj, tests_to_run=test_name, raise_exceptions=True)


@pytest.mark.parametrize("Detector", ALL_DETECTORS)
def test_detector_fit(Detector):
    """Test fit method output."""
    detector = Detector.create_test_instance()
    x = make_annotation_problem(n_timepoints=50, estimator_type="None")
    y = pd.Series(np.zeros(len(x)))  # For coverage testing.
    fit_detector = detector.fit(x, y)
    assert issubclass(detector.__class__, BaseDetector)
    assert issubclass(fit_detector.__class__, BaseDetector)
    assert isinstance(fit_detector, Detector)


@pytest.mark.parametrize("Detector", ALL_DETECTORS)
def test_detector_predict(Detector):
    """Test predict method output."""
    detector = Detector.create_test_instance()
    x = generate_anomalous_data(means=10, random_state=60)
    y = detector.fit_predict(x)
    assert isinstance(y, (pd.Series, pd.DataFrame))


@pytest.mark.parametrize("Detector", ALL_DETECTORS)
def test_detector_transform(Detector):
    """Test transform method output."""
    detector = Detector.create_test_instance()
    x = generate_anomalous_data(means=10, random_state=61)
    y = detector.fit_transform(x)
    assert isinstance(y, (pd.Series, pd.DataFrame))
    assert len(x) == len(y)


@pytest.mark.parametrize("Detector", ALL_DETECTORS)
def test_detector_score_transform(Detector):
    """Test score_transform method output."""
    detector = Detector.create_test_instance()
    x = generate_anomalous_data(means=10, random_state=62)
    try:
        y = detector.fit(x).score_transform(x)
        assert isinstance(y, (pd.Series, pd.DataFrame))
    except NotImplementedError:
        pass


@pytest.mark.parametrize("Detector", ALL_DETECTORS)
def test_detector_update(Detector):
    """Test update method output."""
    detector = Detector.create_test_instance()
    x = make_annotation_problem(n_timepoints=30, estimator_type="None")
    x_train = x[:20].to_frame()
    x_next = x[20:].to_frame()
    detector.fit(x_train)
    detector.update_predict(x_next)
    assert issubclass(detector.__class__, BaseDetector)
    assert isinstance(detector, Detector)


def test_detector_not_implemented_methods():
    detector = BaseDetector()
    x = make_annotation_problem(n_timepoints=20, estimator_type="None")
    with pytest.raises(NotImplementedError):
        detector.fit(x)

    detector._is_fitted = True  # Required for the following functions to run
    with pytest.raises(NotImplementedError):
        detector.predict(x)
    with pytest.raises(NotImplementedError):
        detector.transform(x)
    with pytest.raises(NotImplementedError):
        detector.score_transform(x)
    with pytest.raises(NotImplementedError):
        detector.dense_to_sparse(x)
    with pytest.raises(NotImplementedError):
        detector.sparse_to_dense(x, x.index, pd.Index(["a"]))



"""Change detection algorithms."""

from skchange.change_detectors.base import ChangeDetector
from skchange.change_detectors.moscore import Moscore
from skchange.change_detectors.pelt import Pelt
from skchange.change_detectors.seeded_binseg import SeededBinarySegmentation

BASE_CHANGE_DETECTORS = [ChangeDetector]
CHANGE_DETECTORS = [Moscore, Pelt, SeededBinarySegmentation]

__all__ = BASE_CHANGE_DETECTORS + CHANGE_DETECTORS



"""Tests for CAPA and all available savings."""

import pandas as pd
import pytest

from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.mvcapa import Mvcapa
from skchange.costs.saving_factory import VALID_SAVINGS
from skchange.datasets.generate import generate_teeth_data


@pytest.mark.parametrize("saving", VALID_SAVINGS)
@pytest.mark.parametrize("detector_class", [Capa, Mvcapa])
def test_capa_anomalies(detector_class, saving):
    """Test Capa anomalies."""
    n_segments = 2
    seg_len = 20
    df = generate_teeth_data(
        n_segments=n_segments,
        mean=20,
        segment_length=seg_len,
        p=5,
        affected_proportion=0.2,
        random_state=8,
    )
    detector = detector_class(
        saving=saving,
        collective_penalty_scale=2.0,
        ignore_point_anomalies=True,  # To get test coverage.
    )
    anomalies = detector.fit_predict(df)
    if isinstance(anomalies, pd.DataFrame):
        anomalies = anomalies.iloc[:, 0]
    # End point also included as a changepoint
    assert (
        len(anomalies) == 1
        and anomalies.array.left[0] == seg_len
        and anomalies.array.right[0] == 2 * seg_len - 1
    )



"""Base classes for anomaly detectors."""

import numpy as np
import pandas as pd

from skchange.base import BaseDetector


class PointAnomalyDetector(BaseDetector):
    """Base class for anomaly detectors.

    Anomaly detectors detect individual data points that are considered anomalous.

    Output format of the predict method: See the dense_to_sparse method.
    Output format of the transform method: See the sparse_to_dense method.

    Needs to be implemented:
    - _fit(self, X, y=None) -> self
    - _predict(self, X) -> pd.Series

    Optional to implement:
    - _score_transform(self, X) -> pd.Series
    - _update(self, X, y=None) -> self
    """

    @staticmethod
    def sparse_to_dense(
        y_sparse: pd.Series, index: pd.Index, columns: pd.Index = None
    ) -> pd.Series:
        """Convert the sparse output from the predict method to a dense format.

        Parameters
        ----------
        y_sparse : pd.Series
            The sparse output from an anomaly detector's predict method.
        index : array-like
            Indices that are to be annotated according to ``y_sparse``.
        columns: array-like
            Not used. Only for API compatibility.

        Returns
        -------
        pd.Series where 0-entries are normal and 1-entries are anomalous.
        """
        y_dense = pd.Series(0, index=index, name="anomaly_label", dtype="int64")
        y_dense.iloc[y_sparse.values] = 1
        return y_dense

    @staticmethod
    def dense_to_sparse(y_dense: pd.Series) -> pd.Series:
        """Convert the dense output from the transform method to a sparse format.

        Parameters
        ----------
        y_dense : pd.Series
            The dense output from an anomaly detector's transform method.
            0-entries are normal and >0-entries are anomalous.

        Returns
        -------
        pd.Series of the integer locations of the anomalous data points.
        """
        # The sparse format only uses integer positions, so we reset the index.
        y_dense = y_dense.reset_index(drop=True)

        anomalies = y_dense.iloc[y_dense.values > 0].index
        return PointAnomalyDetector._format_sparse_output(anomalies)

    @staticmethod
    def _format_sparse_output(anomalies) -> pd.Series:
        """Format the sparse output of anomaly detectors.

        Can be reused by subclasses to format the output of the _predict method.
        """
        return pd.Series(anomalies, name="anomaly", dtype="int64")


class CollectiveAnomalyDetector(BaseDetector):
    """Base class for collective anomaly detectors.

    Collective anomaly detectors detect segments of data points that are considered
    anomalous.

    Output format of the predict method: See the dense_to_sparse method.
    Output format of the transform method: See the sparse_to_dense method.

    Needs to be implemented:
    - _fit(self, X, y=None) -> self
    - _predict(self, X) -> pd.Series

    Optional to implement:
    - _score_transform(self, X) -> pd.Series
    - _update(self, X, y=None) -> self
    """

    @staticmethod
    def sparse_to_dense(
        y_sparse: pd.Series, index: pd.Index, columns: pd.Index = None
    ) -> pd.Series:
        """Convert the sparse output from the predict method to a dense format.

        Parameters
        ----------
        y_sparse : pd.Series[pd.Interval]
            The collective anomaly intervals.
        index : array-like
            Indices that are to be annotated according to ``y_sparse``.
        columns: array-like
            Not used. Only for API compatibility.

        Returns
        -------
        pd.Series where 0-entries are normal and each collective anomaly are labelled
            from 1, ..., K.
        """
        labels = pd.IntervalIndex(y_sparse).get_indexer(index)
        # get_indexer return values 0 for the values inside the first interval, 1 to
        # the values within the next interval and so on, and -1 for values outside any
        # interval. The skchange convention is that 0 is normal and > 0 is anomalous,
        # so we add 1 to the result.
        labels += 1
        return pd.Series(labels, index=index, name="anomaly_label", dtype="int64")

    @staticmethod
    def dense_to_sparse(y_dense: pd.Series) -> pd.Series:
        """Convert the dense output from the transform method to a sparse format.

        Parameters
        ----------
        y_dense : pd.Series
            The dense output from a collective anomaly detector's transform method:
            An integer series where 0-entries are normal and each collective anomaly
            are labelled from 1, ..., K.

        Returns
        -------
        pd.Series[pd.Interval] containing the collective anomaly intervals.

        Notes
        -----
        The start and end points of the intervals can be accessed by
        output.array.left and output.array.right, respectively.
        """
        # The sparse format only uses integer positions, so we reset the index.
        y_dense = y_dense.reset_index(drop=True)

        y_anomaly = y_dense.loc[y_dense.values > 0]
        anomaly_locations_diff = y_anomaly.index.diff()

        first_anomaly_start = y_anomaly.index[:1].to_numpy()
        anomaly_starts = y_anomaly.index[anomaly_locations_diff > 1]
        anomaly_starts = np.insert(anomaly_starts, 0, first_anomaly_start)

        last_anomaly_end = y_anomaly.index[-1:].to_numpy()
        anomaly_ends = y_anomaly.index[np.roll(anomaly_locations_diff > 1, -1)]
        anomaly_ends = np.insert(anomaly_ends, len(anomaly_ends), last_anomaly_end)

        anomaly_intervals = list(zip(anomaly_starts, anomaly_ends))
        return CollectiveAnomalyDetector._format_sparse_output(
            anomaly_intervals, closed="both"
        )

    @staticmethod
    def _format_sparse_output(
        anomaly_intervals: list[tuple[int, int]], closed: str = "both"
    ) -> pd.Series:
        """Format the sparse output of collective anomaly detectors.

        Can be reused by subclasses to format the output of the _predict method.
        """
        return pd.Series(
            pd.IntervalIndex.from_tuples(anomaly_intervals, closed=closed),
            name="anomaly_interval",
        )


class SubsetCollectiveAnomalyDetector(BaseDetector):
    """Base class for subset collective anomaly detectors.

    Subset collective anomaly detectors detect segments of multivariate time series data
    that are considered anomalous, and also provide information on which components of
    the data are affected.

    Output format of the predict method: See the dense_to_sparse method.
    Output format of the transform method: See the sparse_to_dense method.

    Output format of the predict method:

    Needs to be implemented:
    - _fit(self, X, y=None) -> self
    - _predict(self, X) -> pd.DataFrame

    Optional to implement:
    - _score_transform(self, X) -> pd.Series
    - _update(self, X, y=None) -> self
    """

    @staticmethod
    def sparse_to_dense(
        y_sparse: pd.DataFrame, index: pd.Index, columns: pd.Index
    ) -> pd.DataFrame:
        """Convert the sparse output from the predict method to a dense format.

        Parameters
        ----------
        y_sparse : pd.DataFrame
            The sparse output from the predict method. The first column must contain the
            anomaly intervals, the second column must contain a list of the affected
            columns.
        index : array-like
            Indices that are to be annotated according to ``y_sparse``.
        columns : array-like
            Columns that are to be annotated according to ``y_sparse``.

        Returns
        -------
        pd.DataFrame where 0-entries are normal and each collective anomaly are labelled
            from 1, ..., K.
        """
        anomaly_intervals = y_sparse.iloc[:, 0].array
        anomaly_starts = anomaly_intervals.left
        anomaly_ends = anomaly_intervals.right
        anomaly_columns = y_sparse.iloc[:, 1]

        start_is_open = anomaly_intervals.closed in ["neither", "right"]
        if start_is_open:
            anomaly_starts += 1  # Exclude the start index in the for loop below.
        end_is_closed = anomaly_intervals.closed in ["both", "right"]
        if end_is_closed:
            anomaly_ends += 1  # Include the end index in the for loop below.

        labels = np.zeros((len(index), len(columns)), dtype="int64")
        anomalies = zip(anomaly_starts, anomaly_ends, anomaly_columns)
        for i, (start, end, affected_columns) in enumerate(anomalies):
            labels[start:end, affected_columns] = i + 1

        return pd.DataFrame(labels, index=index, columns=columns)

    @staticmethod
    def dense_to_sparse(y_dense: pd.DataFrame):
        """Convert the dense output from the transform method to a sparse format.

        Parameters
        ----------
        y_dense : pd.DataFrame
            The dense output from the transform method.

        Returns
        -------
        pd.DataFrame with columns
            anomaly_interval: Intervals of the collective anomalies.
            anomaly_columns: Affected columns of the collective anomalies.
        """
        # The sparse format only uses integer positions, so we reset index and columns.
        y_dense = y_dense.reset_index(drop=True)
        y_dense.columns = range(y_dense.columns.size)

        anomaly_intervals = []
        unique_labels = np.unique(y_dense.values)
        for i in unique_labels[unique_labels > 0]:
            anomaly_mask = y_dense == i
            which_columns = anomaly_mask.any(axis=0)
            which_rows = anomaly_mask.any(axis=1)
            anomaly_columns = anomaly_mask.columns[which_columns].to_list()
            anomaly_start = anomaly_mask.index[which_rows][0]
            anomaly_end = anomaly_mask.index[which_rows][-1]
            anomaly_intervals.append((anomaly_start, anomaly_end, anomaly_columns))

        return SubsetCollectiveAnomalyDetector._format_sparse_output(
            anomaly_intervals, closed="both"
        )

    @staticmethod
    def _format_sparse_output(
        collective_anomalies: list[tuple[int, int, np.ndarray]],
        closed: str = "both",
    ) -> pd.DataFrame:
        """Format the sparse output of subset collective anomaly detectors.

        Can be reused by subclasses to format the output of the _predict method.

        Parameters
        ----------
        collective_anomalies : list
            List of tuples containing start and end indices of collective
            anomalies and a np.array of the affected components/columns.
        closed : str
            Whether the (start, end) tuple correspond to intervals that are closed
            on the left, right, both, or neither.

        Returns
        -------
        pd.DataFrame with columns
            anomaly_interval: Intervals of the collective anomalies.
            anomaly_columns: Affected columns of the collective anomalies.
        """
        anomaly_intervals = [(start, end) for start, end, _ in collective_anomalies]
        affected_components = [components for _, _, components in collective_anomalies]
        return pd.DataFrame(
            {
                "anomaly_interval": pd.IntervalIndex.from_tuples(
                    anomaly_intervals, closed=closed
                ),
                "anomaly_columns": affected_components,
            }
        )



"""Interactive exploration of the Moscore change detector."""

import numpy as np
import plotly.express as px
from numba import njit

from skchange.change_detectors.moscore import Moscore
from skchange.datasets.generate import add_linspace_outliers, generate_teeth_data
from skchange.utils.benchmarking.profiler import Profiler

# Simple univariate example
df = generate_teeth_data(n_segments=2, mean=10, segment_length=100, p=1, random_state=2)
detector = Moscore()
changepoints = detector.fit_predict(df)
labels = detector.transform(df)
scores = detector.score_transform(df)
px.scatter(scores)


# Profiling
n = int(1e6)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = Moscore(bandwidth=50)
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()


# Variance score
df = generate_teeth_data(
    n_segments=2, variance=16, segment_length=100, p=1, random_state=1
)
detector = Moscore(score="mean_var")
changepoints = detector.fit_predict(df)
px.scatter(df)
px.scatter(detector.scores)


# Custom score
@njit
def col_median(X: np.ndarray) -> np.ndarray:
    """Compute the median of each column of X."""
    m = X.shape[1]
    medians = np.zeros(m)
    for j in range(m):
        medians[j] = np.median(X[:, j])
    return medians


@njit
def init_spike_score(X: np.ndarray) -> np.ndarray:
    """Initialize the spike score."""
    return X


@njit
def spike_score(
    precomputed_params: np.ndarray,
    start: np.ndarray,
    end: np.ndarray,
    split: np.ndarray,
) -> float:
    """Calculate the score for a spike at the split point."""
    X = precomputed_params
    baseline_median = np.zeros((len(start), X.shape[1]))
    for i, (s, e) in enumerate(zip(start, end)):
        baseline_median[i] = col_median(X[s : e + 1])
    return np.sum(np.abs(X[split] - baseline_median), axis=1)


df = generate_teeth_data(n_segments=1, mean=0, segment_length=100, p=1)
df = add_linspace_outliers(df, n_outliers=4, outlier_size=10)
score = (spike_score, init_spike_score)
detector = Moscore(score, bandwidth=5)
anomalies = detector.fit_predict(df)
px.scatter(detector.scores)
px.scatter(df)



"""Basic tests for all anomaly detectors."""

import pandas as pd
import pytest

from skchange.anomaly_detectors import ANOMALY_DETECTORS, COLLECTIVE_ANOMALY_DETECTORS
from skchange.anomaly_detectors.base import CollectiveAnomalyDetector
from skchange.datasets.generate import generate_anomalous_data

true_anomalies = [(30, 34), (70, 75)]
anomaly_data = generate_anomalous_data(
    100, anomalies=true_anomalies, means=[10.0, 15.0], random_state=2
)


@pytest.mark.parametrize("Estimator", COLLECTIVE_ANOMALY_DETECTORS)
def test_collective_anomaly_detector_predict(Estimator):
    """Test collective anomaly detector's predict method (sparse output)."""
    detector = Estimator.create_test_instance()
    anomalies = detector.fit_predict(anomaly_data)
    if isinstance(anomalies, pd.DataFrame):
        anomalies = anomalies.iloc[:, 0]

    assert len(anomalies) == len(true_anomalies)
    for i, (start, end) in enumerate(true_anomalies):
        assert anomalies.array.left[i] == start and anomalies.array.right[i] == end


@pytest.mark.parametrize("Estimator", COLLECTIVE_ANOMALY_DETECTORS)
def test_collective_anomaly_detector_transform(Estimator):
    """Test collective anomaly detector's transform method (dense output)."""
    detector = Estimator.create_test_instance()
    labels = detector.fit_transform(anomaly_data)
    if isinstance(labels, pd.DataFrame):
        labels = labels.iloc[:, 0]

    true_collective_anomalies = pd.IntervalIndex.from_tuples(
        true_anomalies, closed="both"
    )
    true_anomaly_labels = CollectiveAnomalyDetector.sparse_to_dense(
        true_collective_anomalies, anomaly_data.index
    )
    labels.equals(true_anomaly_labels)

    # Similar test that does not depend on sparse_to_dense, just to be sure.
    assert labels.nunique() == len(true_anomalies) + 1
    for i, (start, end) in enumerate(true_anomalies):
        assert (labels.iloc[start : end + 1] == i + 1).all()


@pytest.mark.parametrize("Estimator", ANOMALY_DETECTORS)
def test_anomaly_detector_sparse_to_dense(Estimator):
    """Test that predict + sparse_to_dense == transform."""
    detector = Estimator.create_test_instance()
    anomalies = detector.fit_predict(anomaly_data)
    labels_predict_convert = detector.sparse_to_dense(
        anomalies, anomaly_data.index, anomaly_data.columns
    )
    labels_transform = detector.fit_transform(anomaly_data)
    assert labels_predict_convert.equals(labels_transform)


@pytest.mark.parametrize("Estimator", ANOMALY_DETECTORS)
def test_anomaly_detector_dense_to_sparse(Estimator):
    """Test that transform + dense_to_sparse == predict."""
    detector = Estimator.create_test_instance()
    labels = detector.fit_transform(anomaly_data)
    anomalies_transform_convert = detector.dense_to_sparse(labels)
    anomalies_predict = detector.fit_predict(anomaly_data)
    assert anomalies_transform_convert.equals(anomalies_predict)



"""Test statistic for differences in the mean and/or variance for multivariate data."""

__author__ = ["johannvk"]

import numpy as np
from numba import njit

from skchange.utils.numba.stats import log_det_covariance


@njit(cache=True)
def half_integer_digamma(twice_n: int) -> float:
    """Calculate the digamma function for half integer values, i.e. twice_n/2.

    The digamma function is the logarithmic derivative of the gamma function.
    This function is capable of calculating the
    digamma function for half integer values.

    Source: https://en.wikipedia.org/wiki/Digamma_function

    Parameters
    ----------
    twice_n : int
        Twice the integer value.

    Returns
    -------
    res : float
        Value of the digamma function for the half integer value.
    """
    assert twice_n > 0, "n must be a positive integer."

    if twice_n % 2 == 0:
        # Even integer: twice_n = 2n
        res = -np.euler_gamma
        n = twice_n // 2
        for k in range(0, n - 1):
            res += 1.0 / (k + 1.0)
    else:
        res = -2 * np.log(2) - np.euler_gamma
        # Odd integer: twice_n = 2n + 1
        n = (twice_n - 1) // 2
        for k in range(1, n + 1):
            res += 2.0 / (2.0 * k - 1.0)

    return res


@njit(cache=True)
def likelihood_ratio_expected_value(
    sequence_length: int, cut_point: int, dimension: int
) -> float:
    """Calculate the expected value of twice the negative log likelihood ratio.

    We check that the cut point is within the sequence length, and that both 'k' and 'n'
    are large enough relative to the dimension 'p', to ensure that the expected
    value is finite.
    Should at least have 'p+1' points on each side of a split, for 'p' dimensional data.

    Parameters
    ----------
    sequence_length : int
        Length of the sequence.
    cut_point : int
        Cut point of the sequence.
    dimension : int
        Dimension of the data.

    Returns
    -------
    g_k_n : float
        Expected value of twice the negative log likelihood ratio.

    """
    n, k, p = sequence_length, cut_point, dimension

    assert 0 < k < n, "Cut point 'k' must be within the sequence length 'n'."
    assert p > 0, "Dimension 'p' must be a positive integer."
    assert k > (p + 1), "Cut point 'k' must be larger than the dimension + 1."
    assert n - k > (
        p + 1
    ), "Run length after cut point 'n - k' must be larger than dimension + 1."

    g_k_n = p * (
        np.log(2)
        + (n - 1) * np.log(n - 1)
        - (n - k - 1) * np.log(n - k - 1)
        - (k - 1) * np.log(k - 1)
    )

    for j in range(1, p + 1):
        g_k_n += (
            (n - 1) * half_integer_digamma(n - j)
            - (k - 1) * half_integer_digamma(k - j)
            - (n - k - 1) * half_integer_digamma(n - k - j)
        )

    return g_k_n


@njit(cache=True)
def bartlett_correction(twice_negated_log_lr, sequence_length, cut_point, dimension):
    """Calculate the Bartlett correction for the twice negated log likelihood ratio.

    Parameters
    ----------
    twice_negated_log_lr : float
        Twice the negative log likelihood ratio.
    sequence_length : int
        Length of the sequence.
    cut_point : int
        Cut point of the sequence.
    dimension : int
        Dimension of the data.

    Returns
    -------
    bartlett_corr_log_lr : float
    """
    g_k_n = likelihood_ratio_expected_value(
        sequence_length=sequence_length, cut_point=cut_point, dimension=dimension
    )
    bartlett_corr_log_lr = dimension * (dimension + 3.0) * twice_negated_log_lr / g_k_n

    return bartlett_corr_log_lr


@njit(cache=True)
def init_mean_cov_score(X: np.ndarray) -> np.ndarray:
    """Initialize the precision matrix change point detection."""
    return X


@njit(cache=True)
def _mean_cov_log_det_term(X: np.ndarray, start: int, end: int) -> float:
    """Compute the contribution of a segment to the mean_cov score.

    Parameters
    ----------
    X : np.ndarray
        Data matrix. Rows are observations and columns are variables.
    start : int
        Start index of the interval to test for a change in the covariance matrix.
        (Inclusive)
    end : int
        End index of the interval to test for a change in the covariance matrix.
        (Inclusive)

    Returns
    -------
    contribution : float
        Contribution of the segment to the mean and covariance cost.
    """
    log_det_cov = log_det_covariance(X[start : end + 1, :])
    if np.isnan(log_det_cov):
        raise RuntimeError(
            f"The covariance matrix of X[{start}:{end + 1}] is not positive definite."
            + " Quick and dirty fix: Add a tiny amount of random noise to the data."
        )

    return (end - start + 1) * log_det_cov


@njit(cache=True)
def _mean_cov_score(
    X: np.ndarray,
    start: int,
    end: int,
    split: int,
) -> float:
    """Calculate the score for a change in mean and variance.

    The score is computed as the Bartlett corrected log likelihood
    ratio for a test of change in mean and/or variance at the split
    point under a multivariate Gaussian model.

    Parameters
    ----------
    X : np.ndarray
        Data matrix. Rows are observations and columns are variables.
    start : int
        Start index of the interval to test for a change in the precision matrix.
        (Inclusive)
    end : int
        End index of the interval to test for a change in the precision matrix.
        (Inclusive)
    split : int
        Split index of the interval to test for a change in the precision matrix.
        Include the element at the index 'split' in the first segment.

    Returns
    -------
    score : float
        Score (twice negative log likelihood ratio) for the change in mean and variance.
    """
    full_span_loss = _mean_cov_log_det_term(X, start, end)
    pre_split_loss = _mean_cov_log_det_term(X, start, split)
    post_split_loss = _mean_cov_log_det_term(X, split + 1, end)
    twice_negated_log_lr = full_span_loss - pre_split_loss - post_split_loss

    score = bartlett_correction(
        twice_negated_log_lr,
        sequence_length=end + 1 - start,
        cut_point=split - start,
        dimension=X.shape[1],
    )
    return score


@njit(cache=True)
def mean_cov_score(
    precomputed_params: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    splits: np.ndarray,
) -> np.ndarray:
    """Calculate likelihood ratio scores for a change in mean and covariance[1]_.

    References
    ----------
    .. [1] Zamba, K. D., & Hawkins, D. M. (2009). A Multivariate Change-Point Model\
    for Change in Mean Vector and/or Covariance Structure. \
    Journal of Quality Technology, 41(3), 285-303.

    Parameters
    ----------
    precomputed_params : Tuple[np.ndarray, np.ndarray, np.ndarray]
        Precomputed parameters from init_covariance_score.
    starts : np.ndarray
        Start indices of the intervals to test for a change in the covariance.
    ends : np.ndarray
        End indices of the intervals to test for a change in the covariance.
    splits : np.ndarray
        Split indices of the intervals to test for a change in the covariance.

    Returns
    -------
    scores : np.ndarray
        Scores for each start, end, split combination.
    """
    X = precomputed_params
    num_splits = len(splits)

    scores = np.zeros(num_splits, dtype=np.float64)
    for split_idx in range(num_splits):
        scores[split_idx] = _mean_cov_score(
            X, starts[split_idx], ends[split_idx], splits[split_idx]
        )

    return scores



"""Basic tests for all change detectors."""

import pytest

from skchange.change_detectors import CHANGE_DETECTORS
from skchange.datasets.generate import generate_teeth_data

n_segments = 2
seg_len = 50
changepoint_data = generate_teeth_data(
    n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=2
)[0]


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_predict(Estimator):
    """Test changepoint detector predict (sparse output)."""
    detector = Estimator.create_test_instance()
    changepoints = detector.fit_predict(changepoint_data)
    assert len(changepoints) == n_segments - 1 and changepoints[0] == seg_len - 1


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_transform(Estimator):
    """Test changepoint detector transform (dense output)."""
    detector = Estimator.create_test_instance()
    labels = detector.fit_transform(changepoint_data)
    assert labels.nunique() == n_segments
    assert labels[seg_len - 1] == 0.0 and labels[seg_len] == 1.0


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_sparse_to_dense(Estimator):
    """Test that predict + sparse_to_dense == transform."""
    detector = Estimator.create_test_instance()
    changepoints = detector.fit_predict(changepoint_data)
    labels = detector.sparse_to_dense(changepoints, changepoint_data.index)
    labels_transform = detector.fit_transform(changepoint_data)
    assert labels.equals(labels_transform)


@pytest.mark.parametrize("Estimator", CHANGE_DETECTORS)
def test_change_detector_dense_to_sparse(Estimator):
    """Test that transform + dense_to_sparse == predict."""
    detector = Estimator.create_test_instance()
    labels = detector.fit_transform(changepoint_data)
    changepoints = detector.dense_to_sparse(labels)
    changepoints_predict = detector.fit_predict(changepoint_data)
    assert changepoints.equals(changepoints_predict)



"""Detector base class.

    class name: BaseDetector

    Adapted from the BaseSeriesAnnotator class in sktime.

Scitype defining methods:
    fitting                         - fit(self, X, y=None)
    detecting, sparse format        - predict(self, X)
    detecting, dense format         - transform(self, X)
    detection scores, dense         - score_transform(self, X)  [optional]
    updating (temporal)             - update(self, X, y=None)  [optional]

Each detector type (e.g. point anomaly detector, collective anomaly detector,
changepoint detector) are subclasses of BaseDetector (task tag in sktime).
A detector type is defined by the content and format of the output of the predict
method. Each detector type therefore has the following methods for converting between
sparse and dense output formats:
    converting sparse output to dense - sparse_to_dense(y_sparse, index, columns)
    converting dense output to sparse - dense_to_sparse(y_dense)  [optional]

Convenience methods:
    update&detect   - update_predict(self, X)
    fit&detect      - fit_predict(self, X, y=None)
    fit&transform   - fit_transform(self, X, y=None)

Inspection methods:
    hyper-parameter inspection  - get_params()
    fitted parameter inspection - get_fitted_params()

State:
    fitted model/strategy   - by convention, any attributes ending in "_"
    fitted state flag       - check_is_fitted()
"""

__author__ = ["Tveten"]
__all__ = ["BaseDetector"]

import pandas as pd
from sktime.base import BaseEstimator
from sktime.utils.validation.series import check_series


class BaseDetector(BaseEstimator):
    """Base detector.

    An alternative implementation to the BaseSeriesAnnotator class from sktime,
    more focused on the detection of events of interest.
    Safer for now since the annotation module is still experimental.

    All detectors share the common feature that each element of the output from .predict
    indicates the detection of a specific event of interest, such as an anomaly, a
    changepoint, or something else.

    Needs to be implemented:
    - _fit(self, X, y=None) -> self
    - _predict(self, X) -> pd.Series or pd.DataFrame
    - sparse_to_dense(y_sparse, index) -> pd.Series or pd.DataFrame
        * Enables the transform method to work.

    Optional to implement:
    - dense_to_sparse(y_dense) -> pd.Series or pd.DataFrame
    - _score_transform(self, X) -> pd.Series or pd.DataFrame
    - _update(self, X, y=None) -> self
    """

    _tags = {
        "object_type": "detector",  # type of object
        "authors": "Tveten",  # author(s) of the object
        "maintainers": "Tveten",  # current maintainer(s) of the object
    }  # for unit test cases

    def __init__(self):
        self.task = self.get_class_tag("task")
        self.learning_type = self.get_class_tag("learning_type")

        self._is_fitted = False

        self._X = None
        self._y = None

        super().__init__()

    def fit(self, X, y=None):
        """Fit to training data.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Training data to fit model to (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        self :
            Reference to self.

        Notes
        -----
        Creates fitted model that updates attributes ending in "_". Sets
        _is_fitted flag to True.
        """
        X = check_series(X, allow_index_names=True)

        if y is not None:
            y = check_series(y, allow_index_names=True)

        self._X = X
        self._y = y

        # fkiraly: insert checks/conversions here, after PR #1012 I suggest

        self._fit(X=X, y=y)

        # this should happen last
        self._is_fitted = True

        return self

    def _fit(self, X, y=None):
        """Fit to training data.

        core logic

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Training data to fit model to (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        self :
            Reference to self.

        Notes
        -----
        Updates fitted model that updates attributes ending in "_".
        """
        raise NotImplementedError("abstract method")

    def predict(self, X):
        """Detect events in test/deployment data.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Data to detect events in (time series).

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Each element or row corresponds to a detected event. Exact format depends on
            the detector type.
        """
        self.check_is_fitted()

        X = check_series(X, allow_index_names=True)

        # fkiraly: insert checks/conversions here, after PR #1012 I suggest

        y = self._predict(X=X)
        return y

    def _predict(self, X):
        """Detect events in test/deployment data.

        core logic

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Data to detect events in (time series).

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Each element or row corresponds to a detected event. Exact format depends on
            the detector type.
        """
        raise NotImplementedError("abstract method")

    def transform(self, X):
        """Detect events and return the result in a dense format.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Data to detect events in (time series).

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Detections for sequence X. The returned detections will be in the dense
            format, meaning that each element in X will be annotated according to the
            detection results in some meaningful way depending on the detector type.
        """
        y = self.predict(X)
        X_inner = pd.DataFrame(X)
        y_dense = self.sparse_to_dense(y, X_inner.index, X_inner.columns)
        return y_dense

    @staticmethod
    def sparse_to_dense(y_sparse, index, columns=None):
        """Convert the sparse output from a detector to a dense format.

        Parameters
        ----------
        y_sparse : pd.Series
            The sparse output from a detector's predict method. The format of the
            series depends on the task and capability of the annotator.
        index : array-like
            Indices that are to be annotated according to ``y_sparse``.
        columns : array-like, optional
            Columns that are to be annotated according to ``y_sparse``.

        Returns
        -------
        pd.Series or pd.DataFrame of detection labels.
        """
        raise NotImplementedError("abstract method")

    @staticmethod
    def dense_to_sparse(y_dense):
        """Convert the dense output from a detector to a sparse format.

        Parameters
        ----------
        y_dense : pd.Series
            The dense output from a detector's transform method. The format of the
            series depends on the task and capability of the annotator.

        Returns
        -------
        pd.Series
        """
        raise NotImplementedError("abstract method")

    def score_transform(self, X):
        """Return detection scores on test/deployment data.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Data to score (time series).

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Scores for sequence X. Exact format depends on the concrete detector.
        """
        self.check_is_fitted()
        X = check_series(X, allow_index_names=True)
        return self._score_transform(X)

    def _score_transform(self, X):
        """Return detection scores on test/deployment data.

        core logic

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Data to score (time series).

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Scores for sequence X. Exact format depends on the concrete detector.
        """
        raise NotImplementedError("abstract method")

    def update(self, X, y=None):
        """Update model with new data and optional ground truth detections.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Training data to update model with (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        self :
            Reference to self.

        Notes
        -----
        Updates fitted model that updates attributes ending in "_".
        """
        self.check_is_fitted()

        X = check_series(X, allow_index_names=True)

        if y is not None:
            y = check_series(y, allow_index_names=True)

        self._X = X.combine_first(self._X)

        if y is not None:
            self._y = y.combine_first(self._y)

        self._update(X=X, y=y)

        return self

    def _update(self, X, y=None):
        """Update model with new data and optional ground truth detections.

        core logic

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Training data to update model with (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        self :
            Reference to self.

        Notes
        -----
        Updates fitted model that updates attributes ending in "_".
        """
        # default/fallback: re-fit to all data
        self._fit(self._X, self._y)

        return self

    def update_predict(self, X, y=None):
        """Update model with new data and detect events in it.

        Parameters
        ----------
        X : pd.Series, pd.DataFrame or np.ndarray
            Training data to update model with and detect events in (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Each element or row corresponds to a detected event. Exact format depends on
            the detector type.

        Notes
        -----
        Updates fitted model that updates attributes ending in "_".
        """
        self.update(X=X)
        y = self.predict(X=X)

        return y

    def fit_predict(self, X, y=None):
        """Fit to data, then predict it.

        Fits model to X and y with given detector parameters and returns the detected
        events.

        Parameters
        ----------
        X : pd.DataFrame, pd.Series or np.ndarray
            Training data to fit model with and detect events in (time series).
        y : pd.Series, pd.DataFrame or np.ndarray, optional
            Ground truth detections for training if detector is supervised.

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Each element or row corresponds to a detected event. Exact format depends on
            the detector type.
        """
        # Non-optimized default implementation; override when a better
        # method is possible for a given algorithm.
        return self.fit(X, y).predict(X)

    def fit_transform(self, X, y=None):
        """Fit to data, then transform it.

        Fits model to X and y with given detector parameters and returns the detected
        events in a dense format.

        Parameters
        ----------
        X : pd.DataFrame, pd.Series or np.ndarray
            Training data to fit model with and detect events in (time series).
        y : pd.Series or np.ndarray, optional (default=None)
            Target values of data to be detected.

        Returns
        -------
        y : pd.Series or pd.DataFrame
            Detections for sequence X. The returned detections will be in the dense
            format, meaning that each element in X will be annotated according to the
            detection results in some meaningful way depending on the detector type.
        """
        return self.fit(X, y).transform(X)



"""Anomaly detection algorithms."""

from skchange.anomaly_detectors.anomalisers import StatThresholdAnomaliser
from skchange.anomaly_detectors.base import (
    CollectiveAnomalyDetector,
    PointAnomalyDetector,
)
from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.circular_binseg import CircularBinarySegmentation
from skchange.anomaly_detectors.moscore_anomaly import MoscoreAnomaly
from skchange.anomaly_detectors.mvcapa import Mvcapa

BASE_ANOMALY_DETECTORS = [CollectiveAnomalyDetector, PointAnomalyDetector]
COLLECTIVE_ANOMALY_DETECTORS = [
    Capa,
    CircularBinarySegmentation,
    MoscoreAnomaly,
    Mvcapa,
    StatThresholdAnomaliser,
]
POINT_ANOMALY_DETECTORS = []
ANOMALY_DETECTORS = COLLECTIVE_ANOMALY_DETECTORS + POINT_ANOMALY_DETECTORS

__all__ = BASE_ANOMALY_DETECTORS + ANOMALY_DETECTORS
